@@ -31,47 +31,56 @@ namespace netgen
     return hmesh;
   }
   
-
-  Ngx_Mesh :: Ngx_Mesh (shared_ptr<Mesh> amesh) 
-  {
-    if (amesh)
-      mesh = amesh;
-    else
-      mesh = netgen::mesh;
-  }
+  Ngx_Mesh :: Ngx_Mesh (shared_ptr<Mesh> amesh)
+  { mesh = amesh ? amesh : netgen::mesh; }
+  Ngx_Mesh :: Ngx_Mesh (string filename, NgMPI_Comm acomm)
+  { LoadMesh(filename, acomm); }
   
-  Ngx_Mesh * LoadMesh (const string & filename)
+  Ngx_Mesh * LoadMesh (const string & filename, NgMPI_Comm comm)
   {
     netgen::mesh.reset();
-    Ng_LoadMesh (filename.c_str());
+    Ng_LoadMesh (filename.c_str(), comm);
     return new Ngx_Mesh (netgen::mesh);
   }
 
-  void Ngx_Mesh :: LoadMesh (const string & filename)
+  void Ngx_Mesh :: LoadMesh (const string & filename, NgMPI_Comm comm)
   {
     netgen::mesh.reset();
-    Ng_LoadMesh (filename.c_str());
+    Ng_LoadMesh (filename.c_str(), comm);
     // mesh = move(netgen::mesh);
     mesh = netgen::mesh;
   }
 
-  void Ngx_Mesh :: LoadMesh (istream & ist)
+  void Ngx_Mesh :: LoadMesh (istream & ist, NgMPI_Comm comm)
   {
     netgen::mesh = make_shared<Mesh>();
+    netgen::mesh->SetCommunicator(comm);
     netgen::mesh -> Load (ist);
     // mesh = move(netgen::mesh);
     mesh = netgen::mesh;
     SetGlobalMesh (mesh);
   }
 
+  const NgMPI_Comm & Ngx_Mesh :: GetCommunicator() const
+  {
+    // return Valid() ? mesh->GetCommunicator() : NgMPI_Comm{};
+    if (!Valid()) throw Exception("Ngx_mesh::GetCommunicator: don't have a valid mesh");
+    return mesh->GetCommunicator();
+  }
+
   void Ngx_Mesh :: SaveMesh (ostream & ost) const
   {
     mesh -> Save (ost);
   }
 
-  void Ngx_Mesh :: DoArchive (ngstd::Archive & archive)
+  void Ngx_Mesh :: DoArchive (Archive & archive)
   {
-    if (archive.Input()) mesh = make_shared<Mesh>();
+#ifdef PARALLEL
+    if (archive.Input()) {
+      mesh = make_shared<Mesh>();
+      mesh->SetCommunicator(GetCommunicator());
+    }
+#endif
     mesh->DoArchive(archive);
     if (archive.Input())
       {
@@ -125,7 +134,15 @@ namespace netgen
 
   int Ngx_Mesh :: GetNLevels() const
   {
-    return mesh -> mglevels;
+    return max(size_t(1), mesh -> level_nv.Size());
+  }
+
+  size_t Ngx_Mesh :: GetNVLevel(int level) const
+  {
+    if (level >= mesh->level_nv.Size())
+      return mesh->GetNV();
+    else
+      return mesh->level_nv[level];
   }
   
   int Ngx_Mesh :: GetNElements (int dim) const
@@ -675,19 +692,37 @@ namespace netgen
   }
 
 
-  void Ngx_Mesh :: GetParentNodes (int ni, int * parents) const
+  int Ngx_Mesh :: GetHPElementLevel (int ei, int dir) const
   {
-      ni++;
-      if (ni <= mesh->mlbetweennodes.Size())
+    ei++;
+    int level = -1;
+    
+    if (mesh->hpelements)
       {
-          parents[0] = mesh->mlbetweennodes.Get(ni).I1()-1;
-          parents[1] = mesh->mlbetweennodes.Get(ni).I2()-1;
+	int hpelnr = -1;
+	if (mesh->GetDimension() == 2)
+	  hpelnr = mesh->SurfaceElement(ei).GetHpElnr();
+	else
+	  hpelnr = mesh->VolumeElement(ei).GetHpElnr();
+
+        if (hpelnr < 0)
+          throw NgException("Ngx_Mesh::GetHPElementLevel: Wrong hp-element number!");
+        
+        if (dir == 1)
+          level = (*mesh->hpelements)[hpelnr].levelx;
+        else if (dir == 2)
+          level = (*mesh->hpelements)[hpelnr].levely;
+        else if (dir == 3)
+          level = (*mesh->hpelements)[hpelnr].levelz;
+        else
+          throw NgException("Ngx_Mesh::GetHPElementLevel: dir has to be 1, 2 or 3!");
       }
-      else
-          parents[0] = parents[1] = -1;
+    //else
+    //  throw NgException("Ngx_Mesh::GetHPElementLevel only for HPRefinement implemented!");
+
+    return level;	  
   }
-
-
+  
   int Ngx_Mesh :: GetParentElement (int ei) const
   {
       ei++;
@@ -720,33 +755,47 @@ namespace netgen
       return -1;
   }
 
+  int Ngx_Mesh :: GetNIdentifications () const
+  {
+    return mesh->GetIdentifications().GetMaxNr();
+  }
+
+  int Ngx_Mesh :: GetIdentificationType(int idnr) const
+  {
+    return mesh->GetIdentifications().GetType(idnr+1);
+  }
+
+  Ng_BufferMS<int,4> Ngx_Mesh::GetFaceEdges (int fnr) const
+  {
+    const MeshTopology & topology = mesh->GetTopology();
+    NgArrayMem<int,4> ia;
+    topology.GetFaceEdges (fnr+1, ia);
+    Ng_BufferMS<int,4> res(ia.Size());
+    for (size_t i = 0; i < ia.Size(); i++)
+      res[i] = ia[i]-1;
+    return res;
+  }
 
 
 
 
-#ifdef __SSE__
-#include <immintrin.h>
-  
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<1,1> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
     cout << "multi-eltrafo simd called, 1,1,simd" << endl;
   }
 
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<2,2> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
     mesh->GetCurvedElements().CalcMultiPointSurfaceTransformation<2>
-      (elnr, npts,
-       reinterpret_cast<const SIMD<double>*> (xi), sxi,
-       reinterpret_cast<SIMD<double>*> (x), sx,
-       reinterpret_cast<SIMD<double>*> (dxdxi), sdxdxi);
+      (elnr, npts, xi, sxi, x, sx, dxdxi, sdxdxi);
     /*
     for (int i = 0; i < npts; i++)
       {
@@ -773,15 +822,15 @@ namespace netgen
 
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<3,3> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
     mesh->GetCurvedElements().CalcMultiPointElementTransformation
       (elnr, npts,
-       reinterpret_cast<const SIMD<double>*> (xi), sxi,
-       reinterpret_cast<SIMD<double>*> (x), sx,
-       reinterpret_cast<SIMD<double>*> (dxdxi), sdxdxi);
+       xi, sxi,
+       x, sx,
+       dxdxi, sdxdxi);
     /*
     for (int i = 0; i < npts; i++)
       {
@@ -808,33 +857,43 @@ namespace netgen
 
   template<> DLL_HEADER void Ngx_Mesh ::
   MultiElementTransformation<0,2> (int elnr, int npts,
-				   const tAVXd *xi, size_t sxi,
-				   tAVXd * x, size_t sx,
-				   tAVXd * dxdxi, size_t sdxdxi) const
+				   const SIMD<double> *xi, size_t sxi,
+				   SIMD<double> * x, size_t sx,
+				   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
-    cout << "MultiElementtransformation<0,2> simd not implemented" << endl;
+    //cout << "MultiElementtransformation<0,2> simd not implemented" << endl;
+
+    PointIndex pi = mesh->pointelements[elnr].pnum;
+    Point<3> xg = mesh->Point(pi);
+    if (x)
+      for (int j = 0; j < npts; j++)
+	for (int i = 0; i < 2; i++)
+	  x[j*sx+i] = xg(i);
   }
   
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<0,1> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
-    cout << "multi-eltrafo simd called, 0,1,simd" << endl;
+    //cout << "multi-eltrafo simd called, 0,1,simd" << endl;
+    PointIndex pi = mesh->pointelements[elnr].pnum;
+    Point<3> xg = mesh->Point(pi);
+    if (x)
+      for (int j = 0; j < npts; j++)
+	for (int i = 0; i < 1; i++)
+	  x[j*sx+i] = xg(i);
   }
 
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<1,3> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
     mesh->GetCurvedElements().CalcMultiPointSegmentTransformation<3>
-      (elnr, npts,
-       reinterpret_cast<const SIMD<double>*> (xi), sxi,
-       reinterpret_cast<SIMD<double>*> (x), sx,
-       reinterpret_cast<SIMD<double>*> (dxdxi), sdxdxi);
+      (elnr, npts, xi, sxi, x, sx, dxdxi, sdxdxi);
     /*
     double hxi[4][1];
     double hx[4][3];
@@ -857,15 +916,12 @@ namespace netgen
   
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<1,2> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
     mesh->GetCurvedElements().CalcMultiPointSegmentTransformation<2>
-      (elnr, npts,
-       reinterpret_cast<const SIMD<double>*> (xi), sxi,
-       reinterpret_cast<SIMD<double>*> (x), sx,
-       reinterpret_cast<SIMD<double>*> (dxdxi), sdxdxi);
+      (elnr, npts, xi, sxi, x, sx, dxdxi, sdxdxi);
     /*
     for (int i = 0; i < npts; i++)
       {
@@ -892,15 +948,12 @@ namespace netgen
 
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<2,3> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
     mesh->GetCurvedElements().CalcMultiPointSurfaceTransformation<3>
-      (elnr, npts,
-       reinterpret_cast<const SIMD<double>*> (xi), sxi,
-       reinterpret_cast<SIMD<double>*> (x), sx,
-       reinterpret_cast<SIMD<double>*> (dxdxi), sdxdxi);
+      (elnr, npts, xi, sxi, x, sx, dxdxi, sdxdxi);
     /*
     for (int i = 0; i < npts; i++)
       {
@@ -927,9 +980,9 @@ namespace netgen
 
   template<> DLL_HEADER void Ngx_Mesh :: 
   MultiElementTransformation<0,3> (int elnr, int npts,
-                                   const tAVXd * xi, size_t sxi,
-                                   tAVXd * x, size_t sx,
-                                   tAVXd * dxdxi, size_t sdxdxi) const
+                                   const SIMD<double> * xi, size_t sxi,
+                                   SIMD<double> * x, size_t sx,
+                                   SIMD<double> * dxdxi, size_t sdxdxi) const
   {
     for (int i = 0; i < npts; i++)
       {
@@ -948,10 +1001,6 @@ namespace netgen
       }
   }
 
-  
-#endif
-  
-
 
 
   
@@ -964,22 +1013,68 @@ namespace netgen
    int * const indices, int numind) const
 
   {
-    if (mesh->GetDimension() != 1)
-      throw NgException("FindElementOfPoint<1> called for multidim mesh");
-
-    Point<3> p(hp[0], 0,0);
-    for (SegmentIndex si = 0; si < mesh->GetNSeg(); si++)
+    switch (mesh->GetDimension())
       {
-        auto & seg = (*mesh)[si];
-        Point<3> p1 = (*mesh)[seg[0]];
-        Point<3> p2 = (*mesh)[seg[1]];
-        double lam = (p(0)-p1(0)) / (p2(0)-p1(0));
-        if (lam >= -1e-10 && lam <= 1+1e-10)
-          {
-            lami[0] = 1-lam;
-            return si;
-          }
+      case 1:
+        {
+          Point<3> p(hp[0], 0,0);
+          for (SegmentIndex si = 0; si < mesh->GetNSeg(); si++)
+            {
+              auto & seg = (*mesh)[si];
+              Point<3> p1 = (*mesh)[seg[0]];
+              Point<3> p2 = (*mesh)[seg[1]];
+              double lam = (p(0)-p1(0)) / (p2(0)-p1(0));
+              if (lam >= -1e-10 && lam <= 1+1e-10)
+                {
+                  lami[0] = 1-lam;
+                  return si;
+                }
+            }
+        }
+        break;
+      case 2:
+        {
+          Point<3> p(hp[0], hp[1],0);
+          try
+            {
+              auto ind = mesh->GetSurfaceElementOfPoint(p, lami, nullptr,
+                                                        build_searchtree);
+              return ind - 1;
+            }
+          catch(NgException e) // quads not implemented curved yet
+            {
+          for (SegmentIndex si = 0; si < mesh->GetNSeg(); si++)
+            {
+              auto & seg = (*mesh)[si];
+              Point<3> p1 = (*mesh)[seg[0]];
+              Point<3> p2 = (*mesh)[seg[1]];
+              double lam;
+              double r;
+              if (fabs(p2[0]-p1[0]) >= fabs(p2[1]-p1[1]))
+                {
+                  lam = (p[0]-p1[0])/(p2[0]-p1[0]);
+                  r = p[1] - p1[1] - lam*(p2[1]-p1[1]);
+                }
+              else
+                {
+                  lam = (p[1]-p1[1])/(p2[1]-p1[1]);
+                  r = p[0] - p1[0] - lam*(p2[0]-p1[0]);
+                }
+              if ( lam >= -1e-10 && lam <= 1+1e-10 && fabs(r) <= 1e-10 )
+                {
+                  lami[0] = 1-lam;
+                  return si;
+                }
+            }
+            }
+        }
+        break;
+      case 3:
+      default:
+        throw Exception("FindElementOfPoint<1> only implemented for mesh-dimension 1 and 2!");
+        break;
       }
+ 
     return -1;
   }
 
@@ -990,7 +1085,7 @@ namespace netgen
    int * const indices, int numind) const
 
   {
-    Array<int> dummy(numind);
+    NgArray<int> dummy(numind);
     for (int i = 0; i < numind; i++) dummy[i] = indices[i]+1;
     
     double lam3[3];
@@ -1009,7 +1104,7 @@ namespace netgen
     
     if (ind > 0)
       {
-        if(mesh->SurfaceElement(ind).GetType()==QUAD)
+        if(mesh->SurfaceElement(ind).GetType()==QUAD || mesh->SurfaceElement(ind).GetType()==TRIG6)
           {
             lami[0] = lam3[0];
             lami[1] = lam3[1];
@@ -1031,7 +1126,7 @@ namespace netgen
    int * const indices, int numind) const
 
   {
-    Array<int> dummy(numind);
+    NgArray<int> dummy(numind);
     for (int i = 0; i < numind; i++) dummy[i] = indices[i]+1;
     
     Point<3> p3d(p[0], p[1], p[2]);
@@ -1045,10 +1140,35 @@ namespace netgen
     NgLock meshlock (mesh->MajorMutex(), true);
     mesh->BuildCurvedElements(order);
   }
+
+  int Ngx_Mesh :: GetCurveOrder ()
+  {
+    return mesh->GetCurvedElements().GetOrder();
+  }
+
+  void Ngx_Mesh :: EnableTable (string name, bool set)
+  {
+    mesh->GetTopology().EnableTable (name, set);
+    mesh->SetNextTimeStamp();  // update topology will do work
+  }
+
+
   
-  void Ngx_Mesh :: Refine (NG_REFINEMENT_TYPE reftype,
+  template <>
+  DLL_HEADER void Ngx_Mesh :: SetRefinementFlag<2> (size_t elnr, bool flag)
+  {
+    mesh->SurfaceElement(elnr+1).SetRefinementFlag(flag);
+  }
+
+  template <>
+  DLL_HEADER void Ngx_Mesh :: SetRefinementFlag<3> (size_t elnr, bool flag)
+  {
+    mesh->VolumeElement(elnr+1).SetRefinementFlag(flag);    
+  }
+  
+  void Ngx_Mesh :: Refine (NG_REFINEMENT_TYPE reftype, bool onlyonce,
                            void (*task_manager)(function<void(int,int)>),
-                           Tracer tracer)
+                           NgTracer tracer)
   {
     NgLock meshlock (mesh->MajorMutex(), 1);
     
@@ -1056,54 +1176,183 @@ namespace netgen
     biopt.usemarkedelements = 1;
     biopt.refine_p = 0;
     biopt.refine_hp = 0;
+    biopt.onlyonce = onlyonce;
     if (reftype == NG_REFINE_P)
       biopt.refine_p = 1;
     if (reftype == NG_REFINE_HP)
       biopt.refine_hp = 1;
     biopt.task_manager = task_manager;
     biopt.tracer = tracer;
-    
-    const Refinement & ref = mesh->GetGeometry()->GetRefinement();
-    ref.Bisect (*mesh, biopt);
 
+    mesh->GetGeometry()->GetRefinement().Bisect (*mesh, biopt);
     (*tracer)("call updatetop", false);
     mesh -> UpdateTopology(task_manager, tracer);
     (*tracer)("call updatetop", true);
-    mesh -> GetCurvedElements().SetIsHighOrder (false);
+    if(mesh->GetCurvedElements().IsHighOrder())
+      mesh->GetCurvedElements()
+        .BuildCurvedElements(&mesh->GetGeometry()->GetRefinement(),
+                                mesh->GetCurvedElements().GetOrder());
   }
 
 
+
+
+
+  // just copied with redesign
+
+  size_t Ngx_Mesh::GetNP() const
+  {
+    return mesh->GetNP();
+  }
+
+  
+  int Ngx_Mesh::GetSurfaceElementSurfaceNumber (size_t ei) const
+  {
+    if (mesh->GetDimension() == 3)
+      return mesh->GetFaceDescriptor(mesh->SurfaceElement(ei).GetIndex()).SurfNr();
+    else
+      return mesh->LineSegment(ei).si;
+  }
+  int Ngx_Mesh::GetSurfaceElementFDNumber (size_t ei) const
+  {
+    if (mesh->GetDimension() == 3)
+      return mesh->SurfaceElement(ei).GetIndex();
+    else
+      return -1;
+  }
+
+  
+  void Ngx_Mesh::HPRefinement (int levels, double parameter, bool setorders,
+                               bool ref_level)
+  {
+    NgLock meshlock (mesh->MajorMutex(), true);
+    Refinement & ref = const_cast<Refinement&> (mesh->GetGeometry()->GetRefinement());
+    ::netgen::HPRefinement (*mesh, &ref, SPLIT_HP, levels, parameter, setorders, ref_level);
+  }
+
+  void Ngx_Mesh::SplitAlfeld ()
+  {
+    NgLock meshlock (mesh->MajorMutex(), true);
+    Refinement & ref = const_cast<Refinement&> (mesh->GetGeometry()->GetRefinement());
+    ::netgen::HPRefinement (*mesh, &ref, SPLIT_ALFELD, 1, 0.5, true, true);
+  }
+
+  
+int Ngx_Mesh::GetElementOrder (int enr) const
+{
+  if (mesh->GetDimension() == 3)
+    return mesh->VolumeElement(enr).GetOrder();
+  else
+    return mesh->SurfaceElement(enr).GetOrder();
+}
+
+void Ngx_Mesh::GetElementOrders (int enr, int * ox, int * oy, int * oz) const
+{
+  if (mesh->GetDimension() == 3)
+    mesh->VolumeElement(enr).GetOrder(*ox, *oy, *oz);
+  else
+    mesh->SurfaceElement(enr).GetOrder(*ox, *oy, *oz);
+}
+
+void Ngx_Mesh::SetElementOrder (int enr, int order)
+{
+  if (mesh->GetDimension() == 3)
+    return mesh->VolumeElement(enr).SetOrder(order);
+  else
+    return mesh->SurfaceElement(enr).SetOrder(order);
+}
+
+void Ngx_Mesh::SetElementOrders (int enr, int ox, int oy, int oz)
+{
+  if (mesh->GetDimension() == 3)
+    mesh->VolumeElement(enr).SetOrder(ox, oy, oz);
+  else
+    mesh->SurfaceElement(enr).SetOrder(ox, oy);
+}
+
+
+int Ngx_Mesh::GetSurfaceElementOrder (int enr) const
+{
+  return mesh->SurfaceElement(enr).GetOrder();
+}
+
+int Ngx_Mesh::GetClusterRepVertex (int pi) const
+{
+  return mesh->GetClusters().GetVertexRepresentant(pi);
+}
+
+int Ngx_Mesh::GetClusterRepEdge (int pi) const
+{
+  return mesh->GetClusters().GetEdgeRepresentant(pi);
+}
+
+int Ngx_Mesh::GetClusterRepFace (int pi) const
+{
+  return mesh->GetClusters().GetFaceRepresentant(pi);
+}
+
+int Ngx_Mesh::GetClusterRepElement (int pi) const
+{
+  return mesh->GetClusters().GetElementRepresentant(pi);
+}
+
+
+
+
+//HERBERT: falsche Anzahl von Argumenten
+//void Ngx_Mesh::GetSurfaceElementOrders (int enr, int * ox, int * oy, int * oz)
+void Ngx_Mesh::GetSurfaceElementOrders (int enr, int * ox, int * oy) const
+{
+  int d; 
+  mesh->SurfaceElement(enr).GetOrder(*ox, *oy, d);
+}
+
+void Ngx_Mesh::SetSurfaceElementOrder (int enr, int order)
+{
+  return mesh->SurfaceElement(enr).SetOrder(order);
+}
+
+void Ngx_Mesh::SetSurfaceElementOrders (int enr, int ox, int oy)
+{
+  mesh->SurfaceElement(enr).SetOrder(ox, oy);
+}
+
   
 
-#ifdef PARALLEL
+size_t Ngx_Mesh :: GetGlobalVertexNum (int locnum) const
+{
+#ifdef PARALLEL  
+  return mesh->GetParallelTopology().GetGlobalPNum (locnum+1)-1;
+#else
+  return locnum;
+#endif
+}
+
   
-  std::tuple<int,int*>  Ngx_Mesh :: GetDistantProcs (int nodetype, int locnum) const
+FlatArray<int>  Ngx_Mesh :: GetDistantProcs (int nodetype, int locnum) const
   {
+#ifdef PARALLEL
+    if (mesh->GetCommunicator().Size() == 1)
+      return FlatArray<int>(0,nullptr);
     
     switch (nodetype)
       {
       case 0:
-	{
-	  FlatArray<int> dn = mesh->GetParallelTopology().GetDistantPNums(locnum);
-	  return std::tuple<int,int*>(dn.Size(), &dn[0]);
-	}
+        // return mesh->GetParallelTopology().GetDistantPNums(locnum);
+        return mesh->GetParallelTopology().GetDistantProcs(locnum+PointIndex::BASE);
       case 1:
-	{
-	  FlatArray<int> dn = mesh->GetParallelTopology().GetDistantEdgeNums(locnum);
-	  return std::tuple<int,int*>(dn.Size(), &dn[0]);
-	}
+        // return mesh->GetParallelTopology().GetDistantEdgeNums(locnum);
+        return mesh->GetParallelTopology().GetDistantEdgeProcs(locnum);
       case 2:
-	{
-	  FlatArray<int> dn = mesh->GetParallelTopology().GetDistantFaceNums(locnum);
-	  return std::tuple<int,int*>(dn.Size(), &dn[0]);
-	}
+        // return mesh->GetParallelTopology().GetDistantFaceNums(locnum);
+        return mesh->GetParallelTopology().GetDistantFaceProcs(locnum);
       default:
-	return std::tuple<int,int*>(0,nullptr);
+	return FlatArray<int>(0, nullptr);
       }
-  }
-
+#else
+    return FlatArray<int>(0,nullptr);
 #endif
-
+  }
 }
 
 
