@@ -1,6 +1,5 @@
 #include <mystdlib.h>
 #include "meshing.hpp"
-#include "hprefinement.hpp" 
 
 namespace netgen
 {
@@ -22,8 +21,9 @@ namespace netgen
       {
 	pnums[i] = -1;
 	param[i][0] = param[i][1] = param[i][2] = 0;
-        domin=-1; domout=-1; // he:
       }
+    domin=-1; domout=-1; // he:
+    levelx = 0; levely = 0; levelz = 0;
   } 
 
   HPRefElement :: HPRefElement () 
@@ -31,46 +31,40 @@ namespace netgen
     Reset();
   }
 
-  HPRefElement :: HPRefElement(Element & el)
+  HPRefElement :: HPRefElement(Element & el) :
+    type(HP_NONE), index(el.GetIndex()), levelx(0), levely(0), levelz(0), np(el.GetNV()), domin(-1), domout(-1) //domin,out for segments
   { 
     //Reset();
-    np = el.GetNV(); 
     for (int i=0; i<np ; i++) 
       pnums[i] = el[i]; 
     
-    index = el.GetIndex(); 
     const Point3d * points = 
       MeshTopology :: GetVertices (el.GetType());
     for(int i=0;i<np;i++)
       for(int l=0;l<3;l++) 
 	param[i][l] = points[i].X(l+1); 
-    type = HP_NONE; 
-    domin=-1; domout=-1; // he: needed for segments
   }
 
   
-  HPRefElement :: HPRefElement(Element2d & el)
+  HPRefElement :: HPRefElement(Element2d & el) :
+    type(HP_NONE), index(el.GetIndex()), levelx(0), levely(0), levelz(0), np(el.GetNV()), domin(-1), domout(-1) //domin,out for segments
   { 
     //Reset();
-    np = el.GetNV(); 
     
     for (int i=0; i<np ; i++) 
       pnums[i] = el[i]; 
     
-    index = el.GetIndex(); 
     const Point3d * points = 
       MeshTopology :: GetVertices (el.GetType());
     for(int i=0;i<np;i++)
       for(int l=0;l<3;l++) 
 	param[i][l] = points[i].X(l+1); 
-    type = HP_NONE; 
-    domin=-1; domout=-1; // he: needed for segments
   }
 
-  HPRefElement :: HPRefElement(Segment & el)
+  HPRefElement :: HPRefElement(Segment & el) :
+    type(HP_NONE), levelx(0), levely(0), levelz(0), np(2), domin(el.domin), domout(el.domout), singedge_left(el.singedge_left), singedge_right(el.singedge_right)
   { 
     //Reset();
-    np = 2; 
     for (int i=0; i<np ; i++) 
       pnums[i] = el[i];
     const Point3d * points = 
@@ -86,35 +80,18 @@ namespace netgen
       param[i][1] = -1; param[i][2] = -1;
     }
     */
-
-    singedge_left = el.singedge_left; 
-    singedge_right = el.singedge_right; 
-    type = HP_NONE; 
-    // he: needed for orientation!
-    domin = el.domin;
-    domout = el.domout;
   }
   
-  HPRefElement :: HPRefElement(HPRefElement & el)
+  HPRefElement :: HPRefElement(HPRefElement & el) :
+    type(el.type), index(el.index), si(el.si), levelx(el.levelx), levely(el.levely), levelz(el.levelz), np(el.np), coarse_elnr(el.coarse_elnr), domin(el.domin), domout(el.domout), singedge_left(el.singedge_left), singedge_right(el.singedge_right)
+    
   {
     //Reset();
-    np = el.np; 
     for (int i=0; i<np ; i++) 
       {
 	pnums[i] = el[i];
 	for(int l=0; l<3; l++) param[i][l] = el.param[i][l]; 
-      }
-    index = el.index; 
-    levelx = el.levelx; 
-    levely = el.levely; 
-    levelz = el.levelz; 
-    type = el.type; 
-    coarse_elnr = el.coarse_elnr; 
-    singedge_left = el.singedge_left; 
-    singedge_right = el.singedge_right; 
-    domin = el.domin; // he: needed for segments
-    domout=el.domout;
-          
+      }      
   }
 
   void HPRefElement :: SetType (HPREF_ELEMENT_TYPE t) 
@@ -142,12 +119,22 @@ namespace netgen
           param[k][l]=0.;
       }
   }
-  
+
+
+  std::map<HPREF_ELEMENT_TYPE, HPRef_Struct*> & GetHPRegistry()\
+  {
+    static std::map<HPREF_ELEMENT_TYPE, HPRef_Struct*> registry;
+    return registry;
+  }
 
   HPRef_Struct * Get_HPRef_Struct (HPREF_ELEMENT_TYPE type)
   {
     HPRef_Struct * hps = NULL;
-
+    /*
+    if (type >= 100 && type < 1000)
+      if (type != HP_TET_1F_0E_1VA)
+        type = HP_NONETET;
+    */
     switch (type)
       {
       case HP_SEGM:
@@ -196,7 +183,10 @@ namespace netgen
       case HP_TRIG_3SINGEDGES:
 	hps = &reftrig_3singedges; break;
  
- 
+      case HP_TRIG_ALFELD:
+        hps = &reftrig_Alfeld; break;
+
+        
       case HP_QUAD:
 	hps = &refquad; break;
       case HP_DUMMY_QUAD_SINGCORNER:
@@ -309,6 +299,8 @@ namespace netgen
 
       case HP_TET:
 	hps = &reftet; break;
+      case HP_NONETET:
+	hps = &refnonetet; break;
       case HP_TET_0E_1V:
 	hps = &reftet_0e_1v; break;
       case HP_TET_0E_2V:
@@ -395,22 +387,47 @@ namespace netgen
 	hps = &reftet_3ec_1v; break;
       case HP_TET_3EC_2V:
 	hps = &reftet_3ec_2v; break;
+      case HP_TET_3ED_3V:
+	hps = &reftet_3ed_3v; break;
 
 
       case HP_TET_1F_0E_0V:
 	hps = &reftet_1f_0e_0v; break;
-      case HP_TET_1F_0E_1VA:
-	hps = &reftet_1f_0e_1va; break;
+        // case HP_TET_1F_0E_1VA:
+	// hps = &reftet_1f_0e_1va; break;
       case HP_TET_1F_0E_1VB:
 	hps = &reftet_1f_0e_1vb; break;
+      case HP_TET_1F_0E_2V:
+	hps = &reftet_1f_0e_2v; break;
+        // case HP_TET_1F_0E_3V:
+	// hps = &reftet_1f_0e_3v; break;
       case HP_TET_1F_1EA_0V:
 	hps = &reftet_1f_1ea_0v; break;
       case HP_TET_1F_1EB_0V:
 	hps = &reftet_1f_1eb_0v; break;
+      case HP_TET_1F_1E_1VA:
+	hps = &reftet_1f_1e_1va; break;
+      case HP_TET_1F_1E_1VB:
+	hps = &reftet_1f_1e_1vb; break;
+      case HP_TET_1F_1E_2VA:
+	hps = &reftet_1f_1e_2va; break;
+      case HP_TET_1F_1E_2VB:
+	hps = &reftet_1f_1e_2vb; break;
+      case HP_TET_1F_2E_0VA:
+        hps = &reftet_1f_2e_0va; break;
+      case HP_TET_1F_2E_0VB:
+        hps = &reftet_1f_2e_0vb; break;
       case HP_TET_2F_0E_0V:
 	hps = &reftet_2f_0e_0v; break;
-
-
+      case HP_TET_2F_0E_1V:
+	hps = &reftet_2f_0e_1v; break;
+      case HP_TET_2F_1E_0VA:
+        hps = &reftet_2f_1e_0va; break;
+      case HP_TET_2F_1E_0VB:
+        hps = &reftet_2f_1e_0vb; break;
+      case HP_TET_3F_0E_0V:
+        hps = &reftet_3f_0e_0v; break;
+        
       case HP_PRISM:
 	hps = &refprism; break;
       case HP_PRISM_SINGEDGE:
@@ -525,6 +542,8 @@ namespace netgen
 	hps = &refpyramid_0e_1v; break;
       case HP_PYRAMID_EDGES:
 	hps = &refpyramid_edges; break;
+      case HP_PYRAMID_1FB_0E_0V:
+	hps = &refpyramid_1fb_0e_0v; break;
       case HP_PYRAMID_1FB_0E_1VA:
 	hps = &refpyramid_1fb_0e_1va; break;
 
@@ -545,11 +564,21 @@ namespace netgen
       case HP_HEX_1FA_1FB_0E_0V: 
 	hps = &refhex_1fa_1fb_0e_0v; break; 
 
+      case HP_HEX7:
+	hps = &refhex7; break;
+      case HP_HEX7_1FA:
+	hps = &refhex7_1fa; break;
+      case HP_HEX7_1FB:
+	hps = &refhex7_1fb; break;
+       
 
 
       default:
         {
-          hps = NULL;
+          if (GetHPRegistry().count(type))
+            hps = GetHPRegistry()[type];
+          else
+            hps = NULL;
         }
       }
 
@@ -573,13 +602,13 @@ namespace netgen
   }
 
   bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HASHTABLE<int> & edgepoiclt_dom, 
-		       BitArray & cornerpoint, BitArray & edgepoint, INDEX_3_HASHTABLE<int> & faces, INDEX_2_HASHTABLE<int> & face_edges, 
-			INDEX_2_HASHTABLE<int> & surf_edges, Array<int, PointIndex::BASE> & facepoint, int & levels, int & act_ref); 
+		       NgBitArray & cornerpoint, NgBitArray & edgepoint, INDEX_3_HASHTABLE<int> & faces, INDEX_2_HASHTABLE<int> & face_edges, 
+			INDEX_2_HASHTABLE<int> & surf_edges, NgArray<int, PointIndex::BASE> & facepoint, int & levels, int & act_ref); 
 
-  bool ClassifyHPElements (Mesh & mesh, Array<HPRefElement> & elements, int & act_ref, int & levels);
+  bool ClassifyHPElements (Mesh & mesh, NgArray<HPRefElement> & elements, SplittingType split, int & act_ref, int & levels);
   
   
-  void  InitHPElements(Mesh & mesh, Array<HPRefElement> & elements) 
+  void  InitHPElements(Mesh & mesh, NgArray<HPRefElement> & elements) 
   { 
     for(ElementIndex i = 0; i < mesh.GetNE(); i++) 
       {
@@ -622,11 +651,15 @@ namespace netgen
 	HPRefElement hpel(mesh[i]);
 	hpel.coarse_elnr = i; 
 	hpel.type = HP_SEGM; 
-	hpel.index = seg.edgenr + 10000*seg.si; 
+	// hpel.index = seg.edgenr + 10000*seg.si;
+        hpel.index = seg.edgenr;
+        hpel.si = seg.si;
+        /*
 	if(seg.edgenr >= 10000)
 	  {
 	    throw NgException("assumption that seg.edgenr < 10000 is wrong");
 	  }
+        */
 	elements.Append(hpel); 
       }
   }
@@ -634,7 +667,7 @@ namespace netgen
  
  
   /* *******************************  DoRefinement *************************************** */
-  void DoRefinement (Mesh & mesh, Array<HPRefElement> & elements,
+  void DoRefinement (Mesh & mesh, NgArray<HPRefElement> & elements,
 		     Refinement * ref, double fac1) 
   {
     elements.SetAllocSize (5 * elements.Size());
@@ -644,7 +677,7 @@ namespace netgen
     // prepare new points  
     
     fac1 = max(0.001,min(0.33,fac1));
-    cout << " in HP-REFINEMENT with fac1 " << fac1 << endl; 
+    PrintMessage(3, " in HP-REFINEMENT with fac1 ", fac1); 
     *testout << " in HP-REFINEMENT with fac1 " << fac1 <<  endl; 
    
 
@@ -706,8 +739,7 @@ namespace netgen
       {
 	HPRefElement el = elements[i];
 	HPRef_Struct * hprs = Get_HPRef_Struct (el.type);
-	int newlevel = el.levelx + 1;
-
+	int newlevel = el.levelx;
 	int oldnp = 0;
 	switch (hprs->geom)
 	  {
@@ -718,6 +750,7 @@ namespace netgen
 	  case HP_PYRAMID: oldnp = 5; break;
 	  case HP_PRISM: oldnp = 6; break;
 	  case HP_HEX: oldnp = 8; break;
+	  case HP_HEX7: oldnp = 7; break;
             
           default:
             cerr << "HPRefElement: illegal type (3) " << hprs->geom << endl;
@@ -731,6 +764,7 @@ namespace netgen
 	    el.type == HP_TET ||
 	    el.type == HP_PRISM ||
 	    el.type == HP_HEX || 
+	    el.type == HP_HEX7 || 
 	    el.type == HP_PYRAMID)
 	  newlevel = el.levelx;
 
@@ -829,18 +863,29 @@ namespace netgen
 	    HPRefElement newel(el);
 
 	    newel.type = hprs->neweltypes[j]; 
-            
 	    // newel.index = elements[i].index;
 	    // newel.coarse_elnr = elements[i].coarse_elnr;
-	    newel.levelx = newel.levely = newel.levelz = newlevel;
+            if (newel.type == HP_SEGM ||
+                newel.type == HP_TRIG ||
+                newel.type == HP_QUAD ||
+                newel.type == HP_TET ||
+                newel.type == HP_PRISM ||
+                newel.type == HP_HEX || 
+                newel.type == HP_HEX7 || 
+                newel.type == HP_PYRAMID)
+              newel.levelx = newel.levely = newel.levelz = newlevel;
+            else
+              newel.levelx = newel.levely = newel.levelz = newlevel+1;
+	    
             switch(hprsnew->geom) 
 	      {
-	      case HP_SEGM: newel.np=2; break; 
-	      case HP_QUAD: newel.np=4; break; 
-	      case HP_TRIG: newel.np=3; break; 
-	      case HP_HEX: newel.np=8; break; 
-	      case HP_PRISM: newel.np=6; break;
-	      case HP_TET: newel.np=4; break; 
+	      case HP_SEGM:    newel.np=2; break; 
+	      case HP_QUAD:    newel.np=4; break;
+	      case HP_TRIG:    newel.np=3; break;
+	      case HP_HEX:     newel.np=8; break; 
+	      case HP_HEX7:    newel.np=7; break; 
+	      case HP_PRISM:   newel.np=6; break;
+	      case HP_TET:     newel.np=4; break; 
 	      case HP_PYRAMID: newel.np=5; break; 
               default:
                 throw NgException (string("hprefinement.cpp: illegal type"));
@@ -868,7 +913,7 @@ namespace netgen
 	    
 	    if (j == 0) 
 	      elements[i] = newel; // overwrite old element
-	    else        
+	    else
 	      elements.Append (newel);
 	    j++;
 	  }
@@ -882,7 +927,7 @@ namespace netgen
 
   /* ************************** DoRefineDummies ******************************** */
 
-  void DoRefineDummies (Mesh & mesh, Array<HPRefElement> & elements,
+  void DoRefineDummies (Mesh & mesh, NgArray<HPRefElement> & elements,
 			Refinement * ref)
   {
     int oldelsize = elements.Size();
@@ -939,6 +984,7 @@ namespace netgen
 	    for (int k = 0; k < 8; k++)
 	      newel.pnums[k] = newpnums[hprs->newels[j][k]-1];
 	    newel.index = el.index;
+            newel.si = el.si;
 	    newel.coarse_elnr = el.coarse_elnr;
 	    newel.levelx = newel.levely = newel.levelz = newlevel;
 
@@ -961,7 +1007,7 @@ namespace netgen
 
 
 
-  void SubdivideDegeneratedHexes (Mesh & mesh, Array<HPRefElement> & elements, double fac1)
+  void SubdivideDegeneratedHexes (Mesh & mesh, NgArray<HPRefElement> & elements, double fac1)
   {
     int oldne = elements.Size();
     for (int i = 0; i < oldne; i++)
@@ -1002,7 +1048,7 @@ namespace netgen
 
 	      for (int j = 0; j < 6; j++)  
 		{
-		  Array<int> pts;
+		  NgArray<int> pts;
 		  for (int k = 0; k < 4; k++)
 		    {
 		      bool same = 0;
@@ -1100,7 +1146,7 @@ namespace netgen
   }
 
 
-  void CalcStatistics (Array<HPRefElement> & elements)
+  void CalcStatistics (NgArray<HPRefElement> & elements)
   {
     return;
 #ifdef ABC    
@@ -1252,9 +1298,9 @@ namespace netgen
 
 
 
-  void ReorderPoints (Mesh & mesh, Array<HPRefElement> & hpelements)
+  void ReorderPoints (Mesh & mesh, NgArray<HPRefElement> & hpelements)
   {
-    Array<int, 1> map (mesh.GetNP());
+    NgArray<int, 1> map (mesh.GetNP());
     
     for (int i = 1; i <= mesh.GetNP(); i++)
       map[i] = i;
@@ -1292,10 +1338,10 @@ namespace netgen
         // cout << nwrong << " wrong prisms, " << nright << " right prisms" << endl;
       }
 
-    cout << nwrong << " wrong prisms, " << nright << " right prisms" << endl;
+    PrintMessage(3, nwrong, " wrong prisms, ",  nright, " right prisms");
 
 
-    Array<MeshPoint, 1> hpts(mesh.GetNP());
+    NgArray<MeshPoint, 1> hpts(mesh.GetNP());
 
     for (int i = 1; i <= mesh.GetNP(); i++)
       hpts[map[i]] = mesh.Point(i);
@@ -1315,14 +1361,15 @@ namespace netgen
 
   /* ***************************** HPRefinement ********************************** */
 
-  void HPRefinement (Mesh & mesh, Refinement * ref, int levels, double fac1, bool setorders, bool reflevels)
+  void HPRefinement (Mesh & mesh, Refinement * ref, SplittingType split,
+                     int levels, double fac1, bool setorders, bool reflevels)
   {
     PrintMessage (1, "HP Refinement called, levels = ", levels);
 
  
     // NgLock mem_lock (mem_mutex,1);
 
-    mesh.coarsemesh = new Mesh; 
+    mesh.coarsemesh = make_unique<Mesh>(); 
     *mesh.coarsemesh = mesh;
     
     // #ifdef CURVEDELEMS_NEW
@@ -1331,23 +1378,23 @@ namespace netgen
     // #endif
 
 
-    delete mesh.hpelements;
-    mesh.hpelements = new Array<HPRefElement>;
+    // delete mesh.hpelements;
+    mesh.hpelements = make_unique<NgArray<HPRefElement>>();
         
-    Array<HPRefElement> & hpelements = *mesh.hpelements; 
+    NgArray<HPRefElement> & hpelements = *mesh.hpelements; 
         
     InitHPElements(mesh,hpelements); 
-    
-    Array<int> nplevel;
+
+    NgArray<int> nplevel;
     nplevel.Append (mesh.GetNP());
     
     int act_ref=1;
-    bool sing = ClassifyHPElements (mesh,hpelements, act_ref, levels); 
+    bool sing = ClassifyHPElements (mesh,hpelements, split, act_ref, levels); 
 
     sing = true; // iterate at least once
     while(sing) 
       {
-	cout << " Start new hp-refinement: step " <<  act_ref  << endl; 
+	PrintMessage(3, " Start new hp-refinement: step ", act_ref);
 		
 	DoRefinement (mesh, hpelements, ref, fac1); 
 	DoRefineDummies (mesh, hpelements, ref);
@@ -1375,8 +1422,10 @@ namespace netgen
 		    seg[0] = hpel.pnums[0];
 		    seg[1] = hpel.pnums[1];
 		    // NOTE: only for less than 10000 elements (HACK) !!!
-		    seg.edgenr = hpel.index % 10000;
-		    seg.si     = hpel.index / 10000;
+		    // seg.edgenr = hpel.index % 10000;
+		    // seg.si     = hpel.index / 10000;
+                    seg.edgenr = hpel.index;
+                    seg.si = hpel.si;
 
                     /*
                     seg.epgeominfo[0].dist = hpel.param[0][0]; // he: war hpel.param[0][0]
@@ -1399,7 +1448,8 @@ namespace netgen
                     seg.domin = hpel.domin; seg.domout=hpel.domout; // he: needed for segments!
 		    seg.hp_elnr = i;
 		    seg.singedge_left = hpel.singedge_left; 
-		    seg.singedge_right = hpel.singedge_right; 
+		    seg.singedge_right = hpel.singedge_right;
+                    seg.SetCurved (coarseseg.IsCurved());
 		    mesh.AddSegment (seg); 
 		    break;
 		  }
@@ -1410,14 +1460,17 @@ namespace netgen
 		    Element2d el(hpel.np); 
 		    for(int j=0;j<hpel.np;j++) 
 		      el.PNum(j+1) = hpel.pnums[j]; 
-		    el.hp_elnr = i; 
+		    el.SetHpElnr(i);
 		    el.SetIndex(hpel.index);
 		    if(setorders)
 		      el.SetOrder(act_ref+1,act_ref+1,0); 
+                    if((*mesh.coarsemesh)[SurfaceElementIndex{hpel.coarse_elnr}].IsCurved())
+                        el.SetCurved(true);
 		    mesh.AddSurfaceElement(el);
 		    break; 
 		  } 
 		case HP_HEX:
+		case HP_HEX7:
 		case HP_TET:
 		case HP_PRISM:
 		case HP_PYRAMID:
@@ -1426,9 +1479,11 @@ namespace netgen
 		    for(int j=0;j<hpel.np;j++) 
 		      el.PNum(j+1) = hpel.pnums[j]; 
 		    el.SetIndex(hpel.index); 
-		    el.hp_elnr = i; 
+		    el.SetHpElnr(i);
 		    if(setorders)
 		      el.SetOrder(act_ref+1,act_ref+1,act_ref+1);
+                    if((*mesh.coarsemesh)[ElementIndex{hpel.coarse_elnr}].IsCurved())
+                        el.SetCurved(true);
 		    mesh.AddVolumeElement(el); 
 		    break;
 		  } 
@@ -1438,23 +1493,23 @@ namespace netgen
 				 int(Get_HPRef_Struct (hpel.type) -> geom));
 		}
 	  }
-	cout << " Start with Update Topology " << endl; 
+	PrintMessage(5, " Start with Update Topology ");
 	mesh.UpdateTopology();
-	cout << " Mesh Update Topology done " << endl; 
+	PrintMessage(5, " Mesh Update Topology done ");
 
 	act_ref++; 
 	
-	sing = ClassifyHPElements(mesh,hpelements, act_ref, levels); 
+	sing = ClassifyHPElements(mesh,hpelements, split, act_ref, levels); 
       }
 
-    cout << " HP-Refinement done with " << --act_ref << " refinement steps." << endl; 
+    PrintMessage(3, " HP-Refinement done with ", --act_ref, " refinement steps.");
 
     if(act_ref>=1)
       { 
 	for(ElementIndex i=0;i<mesh.GetNE(); i++) 
 	  { 
 	    // Element el = mesh[i] ;
-	    HPRefElement & hpel = hpelements[mesh[i].hp_elnr];
+	    HPRefElement & hpel = hpelements[mesh[i].GetHpElnr()];
 	    const ELEMENT_EDGE * edges = MeshTopology::GetEdges1 (mesh[i].GetType());
 	    double dist[3] = {0,0,0}; 
 	    int ord_dir[3] = {0,0,0}; 
@@ -1475,7 +1530,10 @@ namespace netgen
 		for(int l=6;l<9;l++) edge_dir[l] = 2; 
 		ord_dir[2] = 2; 
 		ned = 9; 
-		break; 
+		break;
+	      case HEX7:
+                // ????
+                break;
 	      case HEX: 
 		/* cout << " HEX " ; 
 		for(int k=0;k<8;k++) cout << el[k] << "\t" ; 
@@ -1526,7 +1584,7 @@ namespace netgen
 	for(SurfaceElementIndex i=0;i<mesh.GetNSE(); i++) 
 	  { 
 	    // Element2d el = mesh[i] ;
-	    HPRefElement & hpel = hpelements[mesh[i].hp_elnr];
+	    HPRefElement & hpel = hpelements[mesh[i].GetHpElnr()];
 	    const ELEMENT_EDGE * edges = MeshTopology::GetEdges1 (mesh[i].GetType());
 	    double dist[3] = {0,0,0}; 
 	    int ord_dir[3] = {0,0,0}; 
@@ -1572,16 +1630,16 @@ namespace netgen
   }
 
 bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HASHTABLE<int> & edgepoint_dom, 
-		       BitArray & cornerpoint, BitArray & edgepoint, INDEX_3_HASHTABLE<int> & faces, INDEX_2_HASHTABLE<int> & face_edges, 
-			INDEX_2_HASHTABLE<int> & surf_edges, Array<int, PointIndex::BASE> & facepoint, int & levels, int & act_ref)
-{ 
+		       NgBitArray & cornerpoint, NgBitArray & edgepoint, INDEX_3_HASHTABLE<int> & faces, INDEX_2_HASHTABLE<int> & face_edges, 
+			INDEX_2_HASHTABLE<int> & surf_edges, NgArray<int, PointIndex::BASE> & facepoint, int & levels, int & act_ref)
+{
   bool sing = 0; 
   if (mesh.GetDimension() == 3)
     {
 	/*
 	// check, if point has as least 3 different surfs:
 
-	Array<INDEX_3, PointIndex::BASE> surfonpoint(mesh.GetNP());
+	NgArray<INDEX_3, PointIndex::BASE> surfonpoint(mesh.GetNP());
   	surfonpoint = INDEX_3(0,0,0);
 
 	for (SurfaceElementIndex sei = 0; sei < mesh.GetNSE(); sei++)
@@ -1613,7 +1671,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 		sing = 1; 
 	      } 
 	  }
-	cout << endl; 
+	// cout << endl; 
 
 	for (int i = 1; i <= mesh.GetNSeg(); i++)
 	  if (mesh.LineSegment(i).singedge_left * levels >= act_ref)
@@ -1640,7 +1698,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    }
 
 	// if 2 adjacent edges of an element are singular, the 
-	// commen point must be a singular point
+	// common point must be a singular point
 	for (int i = 1; i <= mesh.GetNE(); i++)
 	  {
 	    const Element & el = mesh.VolumeElement(i);
@@ -1705,6 +1763,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 		i3 = INDEX_3(i4.I1(), i4.I2(), i4.I3());
 	      }
 	    faces.Set (i3, domnr);
+            *testout << "set face " << i3 << ", domnr = " << domnr << endl;
 	
 	    for (int j = 0; j < el.GetNP(); j++)
 	      {
@@ -1716,15 +1775,16 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	      }
 	   
 	  }
-	(*testout) << "singular faces = " << faces << endl;
-	(*testout) << "singular faces_edges = " << face_edges << endl;
+	// (*testout) << "singular edges = " << edges << endl;        
+	// (*testout) << "singular faces = " << faces << endl;
+	// (*testout) << "singular faces_edges = " << face_edges << endl;
       }
     else
       {
 	// 2D case
 
 	// check, if point has as least 3 different surfs:
-	Array<INDEX_3, PointIndex::BASE> surfonpoint(mesh.GetNP());
+	NgArray<INDEX_3, PointIndex::BASE> surfonpoint(mesh.GetNP());
 
 	for (int i = 1; i <= mesh.GetNP(); i++)
 	  surfonpoint.Elem(i) = INDEX_3(0,0,0);
@@ -1736,7 +1796,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    
 	    if (seg.singedge_left * levels >= act_ref)
 	      {
-		INDEX_2 i2 (mesh.LineSegment(i)[0], 
+		INDEX_2 i2 = INDEX_2::Sort(mesh.LineSegment(i)[0], 
 			    mesh.LineSegment(i)[1]);
 		edges.Set(i2,1); 
 		edgepoint.Set(i2.I1());
@@ -1752,7 +1812,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    
 	    if (seg.singedge_right * levels >= act_ref)
 	      {
-		INDEX_2 i2 (mesh.LineSegment(i)[1], 
+		INDEX_2 i2 = INDEX_2::Sort(mesh.LineSegment(i)[1], 
 			    mesh.LineSegment(i)[0]);  
 		edges.Set (i2, 1);
 		edgepoint.Set(i2.I1());
@@ -1816,21 +1876,21 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
       }
 
     if (!sing)
-      cout << "PrepareElements no more to do for actual refinement " << act_ref << endl; 
+      PrintMessage(3, "PrepareElements no more to do for actual refinement ", act_ref);
 
     return(sing); 
 }
 
 
 
-  bool ClassifyHPElements (Mesh & mesh, Array<HPRefElement> & elements, int & act_ref, int & levels)
+  bool ClassifyHPElements (Mesh & mesh, NgArray<HPRefElement> & elements, SplittingType split, int & act_ref, int & levels)
   {
     INDEX_2_HASHTABLE<int> edges(mesh.GetNSeg()+1);
-    BitArray edgepoint(mesh.GetNP());
+    NgBitArray edgepoint(mesh.GetNP());
     INDEX_2_HASHTABLE<int> edgepoint_dom(mesh.GetNSeg()+1);
 
     edgepoint.Clear();
-    BitArray cornerpoint(mesh.GetNP());
+    NgBitArray cornerpoint(mesh.GetNP());
     cornerpoint.Clear();
 
     // value = nr > 0 ... refine elements in domain nr
@@ -1838,40 +1898,47 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
     INDEX_3_HASHTABLE<int> faces(mesh.GetNSE()+1);
     INDEX_2_HASHTABLE<int> face_edges(mesh.GetNSE()+1);
     INDEX_2_HASHTABLE<int> surf_edges(mesh.GetNSE()+1);
-    Array<int, PointIndex::BASE> facepoint(mesh.GetNP());
+    NgArray<int, PointIndex::BASE> facepoint(mesh.GetNP());
 
     bool sing = CheckSingularities(mesh, edges, edgepoint_dom, 
 			      cornerpoint, edgepoint, faces, face_edges, 
-			      surf_edges, facepoint, levels, act_ref); 
-  
+			      surf_edges, facepoint, levels, act_ref);
+
+    if (act_ref == 1 && split == SPLIT_ALFELD)
+      sing = true;   
     if(sing==0) return(sing); 
 
     int cnt_undef = 0, cnt_nonimplement = 0;
-    Array<int> misses(10000);
+    NgArray<int> misses(10000);
     misses = 0;
 
     (*testout) << "edgepoint_dom = " << endl << edgepoint_dom << endl;
 
+    
     for( int i = 0; i<elements.Size(); i++) 
       {
-	// *testout << "classify element " << i << endl;
-
 	HPRefElement & hpel = elements[i]; 
 	HPRef_Struct * hprs = Get_HPRef_Struct (hpel.type);
 	HPRefElement old_el = elements[i]; 
 	int dd=3; 
 
-
 	if(act_ref !=1 && (hpel.type == HP_HEX || hpel.type == HP_PRISM || hpel.type == HP_TET 
 			   || hpel.type == HP_PYRAMID || hpel.type == HP_QUAD || hpel.type == HP_TRIG || hpel.type == HP_SEGM)) 
 	  continue; 
 	
-	sing = 1; 
+	sing = 1;
 	switch (hprs->geom)
 	  {
 	  case HP_TET:
 	    {
-	      hpel.type = ClassifyTet(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, faces,face_edges, surf_edges, facepoint); 
+	      hpel.type = ClassifyTet(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, faces,face_edges, surf_edges, facepoint);
+              /*
+              // if (i != 182)
+              if ( (!hpel.PNums().Contains(40)) || (!hpel.PNums().Contains(41))  )
+                hpel.type = HP_NONETET;
+              else
+                cout << "type = " << hpel.type << endl;
+              */
 	      break;
 	    }
 	  case HP_PRISM:
@@ -1882,6 +1949,12 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    
 	      break;
 	    }
+	  case HP_HEX7:
+	    { 
+	      hpel.type = ClassifyHex7(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, faces,
+                                       face_edges, surf_edges, facepoint); 	    	    
+	      break; 
+	    } 
 	  case HP_HEX:
 	    { 
 	      hpel.type = ClassifyHex(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, faces,
@@ -1889,23 +1962,25 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	      break; 
 	    } 
 	  case HP_TRIG: 
-	    { 
+	    {
 	      int dim = mesh.GetDimension(); 
-	      const FaceDescriptor & fd = mesh.GetFaceDescriptor (hpel.GetIndex());	
-	      
-	      hpel.type = ClassifyTrig(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, 
-				       faces, face_edges, surf_edges, facepoint, dim, fd);    
-	     
-	      dd = 2; 
+	      const FaceDescriptor & fd = mesh.GetFaceDescriptor (hpel.GetIndex());
+
+              if (split == SPLIT_HP)
+                hpel.type = ClassifyTrig(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, 
+                                         faces, face_edges, surf_edges, facepoint, dim, fd);    
+              else
+                hpel.type = HP_TRIG_ALFELD;
+
+	      dd = 2;
 	      break; 
 	    } 
 	  case HP_QUAD: 
 	    { 
 	      int dim = mesh.GetDimension(); 
-	      const FaceDescriptor & fd = mesh.GetFaceDescriptor (hpel.GetIndex());	
+	      const FaceDescriptor & fd = mesh.GetFaceDescriptor (hpel.GetIndex());
 	      hpel.type = ClassifyQuad(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, 
 				  faces, face_edges, surf_edges, facepoint, dim, fd);    
-
 	      dd = 2; 
 	      break; 
 	    }
@@ -1920,8 +1995,6 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    {
 	      hpel.type = ClassifyPyramid(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, faces,
 						  face_edges, surf_edges, facepoint); 	    	    
-	      
-	      cout << " ** Pyramid classified  " << hpel.type << endl; 
 	      break; 
 	    }
 	  default:
@@ -1930,15 +2003,13 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	      throw NgException ("hprefinement.cpp: don't know how to set parameters");
 	    }
 	  }
-	    
+
 	if(hpel.type == HP_NONE) 
 	  cnt_undef++; 
 
 	//else 
 	//cout << "elem " << i << " classified type " << hpel.type << endl; 
 
-	
-	
 	if (!Get_HPRef_Struct (hpel.type)) 
 	  {
 	    (*testout) << "hp-element-type " << hpel.type << " not implemented   " << endl;
@@ -1961,15 +2032,13 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	  } 
 
       }
-    
-    
-    cout << "undefined elements update classification: " << cnt_undef << endl;
-    cout << "non-implemented in update classification: " << cnt_nonimplement << endl;
 
+    PrintMessage(3, "undefined elements update classification: ", cnt_undef);
+    PrintMessage(3, "non-implemented in update classification: ", cnt_nonimplement);
+    
     for (int i = 0; i < misses.Size(); i++)
       if (misses[i])
 	cout << " in update classification missing case " << i << " occurred " << misses[i] << " times" << endl;
-
     return(sing); 
   }
 }
